/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2006-2018 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: atf_support.S 1314 2019-01-20 15:22:49Z ertl-honda $
 */
#define TOPPERS_MACRO_ONLY
#define TOPPERS_ASM_MACRO
#include "kernel_impl.h"
#include "target_asm.inc"

#define ATF_SAVE_REGS_SIZE  (17 * 8)

    .bss
    AALIGN(8)

/*
 * ATFのSMC実行時に使用されるレジスタの値を保存する領域
 *
 *  X0 :    Function ID
 *  X1-X3 : Parameter
 *  X4-17 : Scratch registers
 */
    AGLOBAL(atf_smc_uses_regs)
ALABEL(atf_smc_uses_regs)
    .skip   ATF_SAVE_REGS_SIZE * TNUM_PRCID


/*
 * [Macro] Save AFT SMC uses regiters
 *
 * Parameter (Registers) :
 *   None
 *
 * Return value (Registers) :
 *   None
 */
    .macro save_atf_regs
    stp   x18, x19, [sp, #-16]!

    /* Get core ID */
    mrs   x18, mpidr_el1
    and   x18, x18, #0xFF
    /* Set save address to x18 */
    mov   x19, #ATF_SAVE_REGS_SIZE
    mul   x19, x18, x19
    ldr   x18, =atf_smc_uses_regs
    add   x18, x18, x19
    /* Save registers */
    stp   x1,  x2,  [x18, #0x00]
    stp   x3,  x4,  [x18, #0x10]
    stp   x5,  x6,  [x18, #0x20]
    stp   x7,  x8,  [x18, #0x30]
    stp   x9,  x10, [x18, #0x40]
    stp   x11, x12, [x18, #0x50]
    stp   x13, x14, [x18, #0x60]
    stp   x15, x16, [x18, #0x70]
    str   x17,      [x18, #0x80]

    ldp   x18, x19, [sp], #16
    .endm

/*
 * [Macro] Restore AFT SMC uses regiters
 *
 * Parameter (Registers) :
 *   None
 *
 * Return value (Registers) :
 *   None
 */
    .macro restore_atf_regs
    stp   x18, x19, [sp, #-16]!

    /* Get core ID */
    mrs   x18, mpidr_el1
    and   x18, x18, #0xFF
    /* Set restore address to x18 */
    mov   x19, #ATF_SAVE_REGS_SIZE
    mul   x19, x18, x19
    ldr   x18, =atf_smc_uses_regs
    add   x18, x18, x19
    /* Restore registers */
    ldp   x1,  x2,  [x18, #0x00]
    ldp   x3,  x4,  [x18, #0x10]
    ldp   x5,  x6,  [x18, #0x20]
    ldp   x7,  x8,  [x18, #0x30]
    ldp   x9,  x10, [x18, #0x40]
    ldp   x11, x12, [x18, #0x50]
    ldp   x13, x14, [x18, #0x60]
    ldp   x15, x16, [x18, #0x70]
    ldr   x17,      [x18, #0x80]

    ldp   x18, x19, [sp], #16
    .endm

    ATEXT
    AALIGN(4)

/*
 * ATFのTrusted OSベクタテーブル
 */
    AGLOBAL(atf_vector_table)
ALABEL(atf_vector_table)
    b start                 /* PSCI CPU_ONでCPUコアON直後に実行されるコールバック */
    b atf_fiq_entry

/*
 * ATFのTrusted OSベクタエントリ
 */
    AGLOBAL(atf_fiq_entry)
ALABEL(atf_fiq_entry)
    /* 復帰アドレスを設定 */
    msr elr_el1, x0
    /* atf_smc_int(), atf_smc_switch()の戻り値を設定 */
    mov x0, xzr
    /* SPD SMCサービルコール実行前のレジスタを復帰 */
    restore_atf_regs
    /* IRQハンドラへジャンプ */
    b cur_spx_irq_handler

/*
 * ATFのNon-secure OS ダミーエントリ
 */
    AGLOBAL(atf_ns_dummy_entry)
ALABEL(atf_ns_dummy_entry)
    b   atf_ns_dummy_entry

/*
 *  SPD SMCサービルコール
 */
/*
 * [C-lang function] SETVCT: Set TOPPERS kernel entry points (fiq vector etc.)
 *
 * Argument :
 *   [In]  X0 - TOPPERS kernel entry points (fiq vector etc.)
 *
 * Return value:
 *   None
 */
    AALIGN(4)
    AGLOBAL(atf_smc_setvct)
ALABEL(atf_smc_setvct)
    /* SPD SMCサービルコール実行前のレジスタを保存 */
    save_atf_regs
    /* Execute SMC Function call */
    mov   x1, x0
    mov   x0, #0x0000   /* ATFSMC_SET_VECTOR */
    movk  x0, #((1 << 15) | (62 << 8)), lsl #16
    dsb   sy
    smc   #0
    /* SPD SMCサービルコール実行前のレジスタを復帰 */
    restore_atf_regs
    ret

/*
 * [C-lang function] INIT:   Notify TOPPERS kernel initialize complete
 *
 * Argument :
 *   None
 *
 * Return value :
 *    0   Success
 *   -1   NS-EL1 OS isn't ready
 */
    AALIGN(4)
    AGLOBAL(atf_smc_init)
ALABEL(atf_smc_init)
    /* SPD SMCサービルコール実行前のレジスタを保存 */
    save_atf_regs
    /* Execute SMC Function call */
    ldr   x1, =atf_ns_dummy_entry
    mov   x0, #0x0001   /* ATFSMC_INIT_COMPLETE */
    movk  x0, #((1 << 15) | (62 << 8)), lsl #16
    dsb   sy
    smc   #0
    ret

/*
 * [C-lang function] CPUON:  Trun on a cpu core
 *
 * Argument :
 *   [In]  X0 - target cpu core (value of mpidr)
 *
 * Return value:
 *   None
 */
    AALIGN(4)
    AGLOBAL(atf_smc_cpuon)
ALABEL(atf_smc_cpuon)
    /* SPD SMCサービルコール実行前のレジスタを保存 */
    save_atf_regs
    /* Execute SMC Function call */
    mov   x3, xzr
    ldr   x2, =atf_ns_dummy_entry
    mov   x1, x0
    mov   x0, #0x0002   /* ATFSMC_CPU_ON */
    movk  x0, #((1 << 15) | (62 << 8)), lsl #16
    dsb   sy
    smc   #0
    /* SPD SMCサービルコール実行前のレジスタを復帰 */
    restore_atf_regs
    ret

/*
 * [C-lang function] SWITCH: Initiates a switch to the opposite world
 *
 * Argument :
 *   None
 *
 * Return value:
 *   0 - Success
 */
    AALIGN(4)
    AGLOBAL(atf_smc_switch)
ALABEL(atf_smc_switch)
    /* SPD SMCサービルコール実行前のレジスタを保存 */
    save_atf_regs
    /* Execute SMC Function call */
    mov   x0, #0x0003   /* ATFSMC_SWITCH_SECURE_STATE */
    movk  x0, #((1 << 15) | (62 << 8)), lsl #16
    dsb   sy
    smc   #0
    ret
