/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2006-2018 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: gic.h 1323 2019-01-26 13:33:40Z ertl-honda $
 */

/*
 *  ARM Generic Interrupt Controller ドライバ
 */

#ifndef TOPPERS_GIC_H
#define TOPPERS_GIC_H

#include <sil.h>

/*
 *  CPU Interface
 *  0x00 0x1f : 32
 *  優先度は，0x0～0xF の16段階．
 *  0x0が最高優先度で，0xFが最低優先度．
 */
#if TOPPERS_GIC_VER == 2
#define GICC_CTLR   (GICC_BASE + 0x00)  /* コントロールレジスタ */
#define GICC_PMR    (GICC_BASE + 0x04)  /* 優先度マスクレジスタ */
#define GICC_BPR    (GICC_BASE + 0x08)  /* バイナリポイントレジスタ */
#define GICC_IAR    (GICC_BASE + 0x0C)  /* Interrupt Acknowledge Register */
#define GICC_EOIR   (GICC_BASE + 0x10)  /* End of Interrupt Register */
#define GICC_RPR    (GICC_BASE + 0x14)  /* Running Interrupt Register */
#define GICC_HPPIR  (GICC_BASE + 0x18)
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
#define ICC_CTLR_EL1_WRITE(x)       Asm("msr S3_0_C12_C12_4, %0"::"r"(x));
#define ICC_CTLR_EL1_READ(x)        Asm("mrs %0, S3_0_C12_C12_4":"=r"(x));
#define ICC_SRE_EL1_WRITE(x)        Asm("msr S3_0_C12_C12_5, %0"::"r"(x));
#define ICC_SRE_EL1_READ(x)         Asm("mrs %0, S3_0_C12_C12_5":"=r"(x));
#define ICC_SRE_EL2_WRITE(x)        Asm("msr S3_4_C12_C9_5, %0"::"r"(x));
#define ICC_SRE_EL2_READ(x)         Asm("mrs %0, S3_4_C12_C9_5":"=r"(x));
#define ICC_SRE_EL3_WRITE(x)        Asm("msr S3_6_C12_C12_5, %0"::"r"(x));
#define ICC_SRE_EL3_READ(x)         Asm("mrs %0, S3_6_C12_C12_5":"=r"(x));
#define ICC_IGRPEN0_EL1_WRITE(x)    Asm("msr S3_0_C12_C12_6, %0"::"r"(x));
#define ICC_IGRPEN0_EL1_READ(x)     Asm("mrs %0, S3_0_C12_C12_6":"=r"(x));
#define ICC_IGRPEN1_EL1_WRITE(x)    Asm("msr S3_0_C12_C12_7, %0"::"r"(x));
#define ICC_IGRPEN1_EL1_READ(x)     Asm("mrs %0, S3_0_C12_C12_7":"=r"(x));
#define ICC_PMR_EL1_WRITE(x)        Asm("msr S3_0_C4_C6_0, %0"::"r"(x));
#define ICC_PMR_EL1_READ(x)         Asm("mrs %0, S3_0_C4_C6_0":"=r"(x));
#define ICC_BPR0_EL1_WRITE(x)       Asm("msr S3_0_C12_C8_3, %0"::"r"(x));
#define ICC_BPR0_EL1_READ(x)        Asm("mrs %0, S3_0_C12_C8_3":"=r"(x));
#define ICC_BPR1_EL1_WRITE(x)       Asm("msr S3_0_C12_C12_3, %0"::"r"(x));
#define ICC_BPR1_EL1_READ(x)        Asm("mrs %0, S3_0_C12_C12_3":"=r"(x));
#define ICC_IAR0_EL1_WRITE(x)       Asm("msr S3_0_C12_C8_0, %0"::"r"(x));
#define ICC_IAR0_EL1_READ(x)        Asm("mrs %0, S3_0_C12_C8_0":"=r"(x));
#define ICC_IAR1_EL1_WRITE(x)       Asm("msr S3_0_C12_C12_0, %0"::"r"(x));
#define ICC_IAR1_EL1_READ(x)        Asm("mrs %0, S3_0_C12_C12_0":"=r"(x));
#define ICC_EOIR0_EL1_WRITE(x)      Asm("msr S3_0_C12_C8_1, %0"::"r"(x));
#define ICC_EOIR0_EL1_READ(x)       Asm("mrs %0, S3_0_C12_C8_1":"=r"(x));
#define ICC_EOIR1_EL1_WRITE(x)      Asm("msr S3_0_C12_C12_1, %0"::"r"(x));
#define ICC_EOIR1_EL1_READ(x)       Asm("mrs %0, S3_0_C12_C12_1":"=r"(x));
#define ICC_RPR_EL1_WRITE(x)        Asm("msr S3_0_C12_C11_3, %0"::"r"(x));
#define ICC_RPR_EL1_READ(x)         Asm("mrs %0, S3_0_C12_C11_3":"=r"(x));
#define ICC_HPPIR0_EL1_WRITE(x)     Asm("msr S3_0_C12_C8_2, %0"::"r"(x));
#define ICC_HPPIR0_EL1_READ(x)      Asm("mrs %0, S3_0_C12_C8_2":"=r"(x));
#define ICC_HPPIR1_EL1_WRITE(x)     Asm("msr S3_0_C12_C12_2, %0"::"r"(x));
#define ICC_HPPIR1_EL1_READ(x)      Asm("mrs %0, S3_0_C12_C12_2":"=r"(x));
#endif /* TOPPERS_GIC_VER == 2 */

#define GICC_CTLR_ENABLE     0x01
#define GICC_CTLR_ENABLEGRP0 0x01
#define GICC_CTLR_ENABLEGRP1 0x02
#define GICC_CTLR_ACKCTL     0x04
#define GICC_CTLR_CBPR       0x10
#define GICC_CTLR_FIQEN      0x08
#define GICC_CTLR_IRQBYPDISGRP1 (1 << 8)
#define GICC_CTLR_FIQBYPDISGRP1 (1 << 7)
#define GICC_CTLR_IRQBYPDISGRP0 (1 << 6)
#define GICC_CTLR_FIQBYPDISGRP0 (1 << 5)

#define GICC_IAR_INTERRUPTID_MASK   0x03ff
#define GICC_IAR_CPUID_MASK         0x1c00

#ifdef TOPPERS_TZ_S
#if GIC_PRI_LEVEL == 32
#define GICC_PMR_OFFSET 0x03
#define GICC_PMR_MASK   0x1f
#elif GIC_PRI_LEVEL == 64
#define GICC_PMR_OFFSET 0x02
#define GICC_PMR_MASK   0x3f
#elif GIC_PRI_LEVEL == 128
#define GICC_PMR_OFFSET 0x01
#define GICC_PMR_MASK   0x7f
#elif GIC_PRI_LEVEL == 256
#define GICC_PMR_OFFSET 0x00
#define GICC_PMR_MASK   0xff
#else /* GIC_PRI_LEVEL == 32 */
#error
#endif /* GIC_PRI_LEVEL == 32 */
#else /* TOPPERS_TZ_S */
#if GIC_PRI_LEVEL == 32
#define GICC_PMR_OFFSET 0x04
#define GICC_PMR_MASK   0x0f
#elif GIC_PRI_LEVEL == 64
#define GICC_PMR_OFFSET 0x03
#define GICC_PMR_MASK   0x1f
#elif GIC_PRI_LEVEL == 128
#define GICC_PMR_OFFSET 0x02
#define GICC_PMR_MASK   0x3f
#elif GIC_PRI_LEVEL == 256
#define GICC_PMR_OFFSET 0x01
#define GICC_PMR_MASK   0x7f
#else /* GIC_PRI_LEVEL == 32 */
#error
#endif /* GIC_PRI_LEVEL == 32 */
#endif /* TOPPERS_TZ_S */

#if (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
/*
 *  SGI
 */
#define ICC_SGI0R_EL1_WRITE(x)      Asm("msr S3_0_C12_C11_7, %0"::"r"(x));
#define ICC_SGI1R_EL1_WRITE(x)      Asm("msr S3_0_C12_C11_5, %0"::"r"(x));

#define BITPOS_ICC_SGInR_AFF3   48
#define BITPOS_ICC_SGInR_RS     44
#define BITPOS_ICC_SGInR_IRM    40
#define BITPOS_ICC_SGInR_AFF2   32
#define BITPOS_ICC_SGInR_INTID  24
#define BITPOS_ICC_SGInR_AFF1   16
#define BITPOS_ICC_SGInR_TLIST  0

/*
 *  Redistributor
 */
#define GICR_CTLR(x)        (gicr_base[x] + GICR_RD_OFFSET + 0x0000)
#define GICR_IIDR(x)        (gicr_base[x] + GICR_RD_OFFSET + 0x0004)
#define GICR_TYPER(x)       (gicr_base[x] + GICR_RD_OFFSET + 0x0008)
#define GICR_STATUSR(x)     (gicr_base[x] + GICR_RD_OFFSET + 0x0010)
#define GICR_WAKER(x)       (gicr_base[x] + GICR_RD_OFFSET + 0x0014)
 #define GICR_ChildrenAsleep (uint32_t)(1 << 2)
 #define GICR_ProcSleep      (uint32_t)(1 << 1)

#define GICR_SETLPIR(x)     (gicr_base[x] + GICR_RD_OFFSET + 0x0040)
#define GICR_CLRLPIR(x)     (gicr_base[x] + GICR_RD_OFFSET + 0x0048)
#define GICR_PROPBASER(x)   (gicr_base[x] + GICR_RD_OFFSET + 0x0070)
#define GICR_PENDBASER(x)   (gicr_base[x] + GICR_RD_OFFSET + 0x0078)
#define GICR_INVLPIR(x)     (gicr_base[x] + GICR_RD_OFFSET + 0x00A0)
#define GICR_INVALLR(x)     (gicr_base[x] + GICR_RD_OFFSET + 0x00B0)
#define GICR_SYNCR(x)       (gicr_base[x] + GICR_RD_OFFSET + 0x00C0)

#define GICR_IGROUPR0(x)    (gicr_base[x] + GICR_SGI_OFFSET + 0x0080)
#define GICR_ISENABLER0(x)  (gicr_base[x] + GICR_SGI_OFFSET + 0x0100)
#define GICR_ICENABLER0(x)  (gicr_base[x] + GICR_SGI_OFFSET + 0x0180)
#define GICR_ISPENDR0(x)    (gicr_base[x] + GICR_SGI_OFFSET + 0x0200)
#define GICR_ICPENDR0(x)    (gicr_base[x] + GICR_SGI_OFFSET + 0x0280)
#define GICR_ISACTIVER0(x)  (gicr_base[x] + GICR_SGI_OFFSET + 0x0300)
#define GICR_ICACTIVER0(x)  (gicr_base[x] + GICR_SGI_OFFSET + 0x0380)
#define GICR_IPRIORITYRn(x) (gicr_base[x] + GICR_SGI_OFFSET + 0x0400)
#define GICR_ICFGRn(x)      (gicr_base[x] + GICR_SGI_OFFSET + 0x0C00)
#define GICR_IGRPMODR0(x)   (gicr_base[x] + GICR_SGI_OFFSET + 0x0D00)
#define GICR_NSACR(x)       (gicr_base[x] + GICR_SGI_OFFSET + 0x0E00)

#endif /* (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4) */

/*
 *  Distributor
 */
#define GICD_CTLR        (GICD_BASE + 0x0000)       /* コントロールレジスタ */
#define GICD_TYPER       (GICD_BASE + 0x0004)       /* Interrupt Controller タイプレジスタ */
#define GICD_IIDR        (GICD_BASE + 0x0008)       /* Distributor Implementer ID */
#define GICD_SETSPI_NSR  (GICD_BASE + 0x0040)       /* Set SPI Register */
#define GICD_CLRSPI_NSR  (GICD_BASE + 0x0048)       /* Clear SPI Register */
#define GICD_SETSPI_SR   (GICD_BASE + 0x0050)       /* Set SPI, Secure Register */
#define GICD_CLRSPI_SR   (GICD_BASE + 0x0058)       /* Clear SPI, Secure Register */
#define GICD_IGROUPRn    (GICD_BASE + 0x0080)       /* 割込みセキュリティ */
#define GICD_IGROUPR(n)  (GICD_BASE + 0x0080 + (n) * 4)
#define GICD_ISENABLERn  (GICD_BASE + 0x0100)       /* 割込みイネーブルセット   */
#define GICD_ISENABLER(n) (GICD_BASE + 0x100 + (n) * 4)
#define GICD_ICENABLERn  (GICD_BASE + 0x0180)       /* 割込みイネーブルクリアー */
#define GICD_ISPENDRn    (GICD_BASE + 0x0200)       /* 割込みセットペンディング */
#define GICD_ISPENDR(n)  (GICD_BASE + 0x200 + (n) * 4)
#define GICD_ICPENDRn    (GICD_BASE + 0x0280)       /* 割込みクリアーペンディング */
#define GICD_ISACTIVERn  (GICD_BASE + 0x0300)       /* 割込みセットアクティブレジスター */
#define GICD_IPRIORITYRn (GICD_BASE + 0x0400)       /* 割込み優先度レジスタ */
#define GICD_IPRIORITYR(n) (GICD_BASE + 0x0400 + (n) * 4)
#define GICD_ITARGETSRn  (GICD_BASE + 0x0800)       /* 割込みターゲットレジスタ/CA9はSPIターゲットレジスタ */
#define GICD_ITARGETSR(n) (GICD_BASE + 0x0800 + (n) * 4)
#define GICD_ICFGRn      (GICD_BASE + 0x0C00)       /* 割込みコンフィギュレーションレジスタ */
#define GICD_IGRPMODRn   (GICD_BASE + 0x0D00)       /* Interrupt Group Modifier Registers */
#define GICD_SGIR        (GICD_BASE + 0x0F00)       /* ソフトウェア割込みレジスタ  */
#define GICD_CPENDSGIRn  (GICD_BASE + 0x0F10)       /* SGI Clear-Pending Registers */
#define GICD_SPENDSGIRn  (GICD_BASE + 0x0F20)       /* SGI Set-Pending Registers */
#define GICD_IROUTERn    (GICD_BASE + 0x6100)       /* Interrupt Routing Registers */

#define GICD_PPIS        (GICD_BASE + 0x0D00)  /* PPIステータス */
#define GICD_SPIS        (GICD_BASE + 0x0D04)  /* SPIステータス */

#define GICD_CTLR_ENABLE 0x01
#define GICD_CTLR_ENABLEGRP0 0x01
#define GICD_CTLR_ENABLEGRP1 0x02

/*
 *  割込み先のプロセッサの指定
 */
#define GICD_ITARGETSRn_CPU0  0x01
#define GICD_ITARGETSRn_CPU1  0x02
#define GICD_ITARGETSRn_CPU2  0x04
#define GICD_ITARGETSRn_CPU3  0x08

#define GICD_SGIR_CPU0  0x01
#define GICD_SGIR_CPU1  0x02
#define GICD_SGIR_CPU2  0x04
#define GICD_SGIR_CPU3  0x08
#define GICD_SGIR_CPUS  0x0f

#define GICD_SGIR_CPU_OFFSET 16

/*
 *  割込み優先度
 */
#define GIC_MAX_PRIORITY  0x00
#define GIC_MIN_PRIORITY  GICC_PMR_MASK

/*
 *  コンフィギュレーションレジスタの設定値
 */
#define GICD_ICFGRn_EDGE     0x03   /* エッジ割込み */
#define GICD_ICFGRn_LEVEL    0x00   /* レベル割込み */
#define GICD_ICFGRn_N_N      0x00   /* N-Nモデル    */
#define GICD_ICFGRn_1_N      0x01   /* 1-Nモデル    */

/*
 *  GICでサポートしている割込み数
 */
#define GIC_TMIN_INTNO      0U

#ifndef GIC_TMAX_INTNO
#define GIC_TMAX_INTNO    255U
#endif /* GIC_TMAX_INTNO */

#ifndef GIC_TNUM_INT
#define GIC_TNUM_INT      256U
#endif /* GIC_TNUM_INT */

/*
 *  グローバル割込みの開始番号
 */
#define TMIN_GLOBAL_INTNO   32U

/*
 *  割込み番号(GICでの番号)
 */
#define GIC_IRQNO_IPI0     0
#define GIC_IRQNO_IPI1     1
#define GIC_IRQNO_IPI2     2
#define GIC_IRQNO_IPI3     3
#define GIC_IRQNO_IPI4     4
#define GIC_IRQNO_IPI5     5
#define GIC_IRQNO_IPI6     6
#define GIC_IRQNO_IPI7     7
#define GIC_IRQNO_IPI8     8
#define GIC_IRQNO_IPI9     9
#define GIC_IRQNO_IPI10   10
#define GIC_IRQNO_IPI11   11
#define GIC_IRQNO_IPI12   12
#define GIC_IRQNO_IPI13   13
#define GIC_IRQNO_IPI14   14
#define GIC_IRQNO_IPI15   15

#ifndef TOPPERS_MACRO_ONLY

/*
 *  GIC CPU Interface 関連のドライバ
 */
#if TOPPERS_GIC_VER == 2
/*
 *  CPUの割込み優先度マスクを設定
 */ 
Inline void
gicc_set_priority(int pri)
{
	sil_wrw_mem((void *)GICC_PMR,
				(pri << GICC_PMR_OFFSET));
}

/*
 *  CPUの割込み優先度マスクを取得
 */ 
Inline int
gicc_current_priority(void)
{
	return (sil_rew_mem((void *)(GICC_PMR)) >>
			GICC_PMR_OFFSET);    
}

/*
 *  割込みハンドラの出口で必要なIRC操作
 */
Inline void
gicc_end_int(uint8_t id)
{
	sil_wrw_mem((void *)(GICC_EOIR), id);
}

/*
 *  GICのプロセッサの割込み優先度のどのビットを使用するか
 */
Inline void
gicc_set_bp(int mask_bit)
{
	sil_wrw_mem((void *)GICC_BPR, mask_bit);
}

/*
 *  ソフトウェア割込みを発行
 */
Inline void
gic_raise_sgi(int cpu, int id)
{
	DATA_SYNC_BARRIER();
	sil_wrw_mem((void *)GICD_SGIR, (cpu << GICD_SGIR_CPU_OFFSET) | id);
	DATA_SYNC_BARRIER();
}

#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)

/*
 *  CPUの割込み優先度マスクを設定
 */ 
Inline void
gicc_set_priority(int pri)
{
	ICC_PMR_EL1_WRITE((pri << GICC_PMR_OFFSET));
}

/*
 *  CPUの割込み優先度マスクを取得
 */ 
Inline int
gicc_current_priority(void)
{
	volatile uint32_t   reg32_val;

	ICC_PMR_EL1_READ(reg32_val);
	return (reg32_val >> GICC_PMR_OFFSET);
}

/*
 *  割込みハンドラの出口で必要なIRC操作
 */
Inline void
gicc_end_int(uint8_t id)
{
	ICC_EOIR0_EL1_WRITE(id);
}

/*
 *  GICのプロセッサの割込み優先度のどのビットを使用するか
 */
Inline void
gicc_set_bp(int mask_bit)
{
#ifdef TOPPERS_TZ_S
	ICC_BPR0_EL1_WRITE(mask_bit);
#else  /* !TOPPERS_TZ_S */
	ICC_BPR1_EL1_WRITE(mask_bit);
#endif /* TOPPERS_TZ_S */
}

/*
 *  ソフトウェア割込みを発行
 */
Inline void
gic_raise_sgi(int cpu, int id)
{
	volatile uint64_t   reg64_val;

	reg64_val = ((uint64_t)id  << BITPOS_ICC_SGInR_INTID) |
	  ((uint64_t)cpu << BITPOS_ICC_SGInR_TLIST);
	DATA_SYNC_BARRIER();
#ifdef TOPPERS_TZ_S
	ICC_SGI0R_EL1_WRITE(reg64_val);
#else  /* !TOPPERS_TZ_S */
	ICC_SGI1R_EL1_WRITE(reg64_val);
#endif /* TOPPERS_TZ_S */
	DATA_SYNC_BARRIER();
}
#endif /* TOPPERS_GIC_VER == 2 */

/*
 *  GICの初期化
 */
extern void gic_init(void);

/*
  *  GIC CPU Interface の終了
 */
extern void gicc_stop(void);

/*
 *  Distributor 関連のドライバ
 */

/*
 *  割込み禁止
 */
extern void gicd_disable_int(uint8_t id);

/*
 *  割込み許可
 */
extern void gicd_enable_int(uint8_t id);

/*
 *  割込みペンディングクリア
 */
extern void gicd_clear_pending(uint8_t id);

/*
 *  割込みペンディングセット
 */
extern void gicd_set_pending(uint8_t id);

/*
 *  割込み要求のチェック
 */
extern bool_t gicd_probe_int(uint8_t id);

/*
 *  割込み設定のセット
 */
extern void gicd_config(uint8_t id, bool_t is_edge, bool_t is_1_n);

/*
 *  割込み優先度のセット
 *  内部表現で渡す． 
 */
extern void gicd_set_priority(uint8_t id, int pri);

/*
 *  割込みターゲットの設定
 *  CPUはORで指定  
 */
extern void gicd_set_target(uint8_t id, uint8_t cpus);

/*
 *  Ditoributor の初期化
 */
extern void gicd_init(void);

/*
 *  Ditoributor の初期化の終了
 */
extern void gicd_stop(void);

/*
 *  例外ハンドラ（gic_support.S）
 */
extern void cur_spx_irq_handler(void);

#endif /* TOPPERS_MACRO_ONLY */
#endif /* TOPPERS_GIC_H */
