/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2006-2019 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: gic_support.S 1323 2019-01-26 13:33:40Z ertl-honda $
 */

/*
 *  GICドライバ アセンブリ言語部
 */

#define TOPPERS_MACRO_ONLY
#define UINT_C(val)     (val)       /* uint_t型の定数を作るマクロ */
#define ULONG_C(val)    (val)       /* ulong_t型の定数を作るマクロ */
#include "kernel_impl.h"
#include "target_asm.inc"
#include "target_ipi.h"
#include "offset.h"


#if (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
#ifdef TOPPERS_TZ_S
#define ICC_IAR_NO		S3_0_C12_C8_0			/* icc_iar0_el1 */
#define ICC_EOIR_NO		S3_0_C12_C8_1			/* icc_eoir0_el1 */
#else  /* !TOPPERS_TZ_S */
#define ICC_IAR_NO		S3_0_C12_C12_0			/* icc_iar1_el1 */
#define ICC_EOIR_NO		S3_0_C12_C12_1			/* icc_eoir1_el1 */
#endif /* TOPPERS_TZ_S */
#endif /* (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4) */

/*
 *  割込みハンドラ
 */
    ATEXT
    AALIGN(2)
    AGLOBAL(cur_spx_irq_handler)
ALABEL(cur_spx_irq_handler)
    /*
     *  スクラッチレジスタをスタックへ保存．
     */
    sub   sp, sp, #(16 * 10)
    stp   x0, x1, [sp, #0x00]
    stp   x2, x3, [sp, #0x10]
    stp   x4, x5, [sp, #0x20]
    stp   x6, x7, [sp, #0x30]
    stp   x8, x9, [sp, #0x40]
    stp   x10, x11, [sp, #0x50]
    stp   x12, x13, [sp, #0x60]
    stp   x14, x15, [sp, #0x70]
    stp   x16, x17, [sp, #0x80]
    stp   x18, x30, [sp, #0x90]

    /*
     *  elr(戻り先), spsrをスタックへ保存．
     */
    mrs   x0, elr_el1
    mrs   x1, spsr_el1
    stp   x0, x1, [sp, #-16]!

    /*
     *  カーネル管理の割込みのみを禁止．
     */
    mov   x0, #(DAIF_CPULOCK AOR DAIF_ALWAYS_SET)
    msr   daif, x0

    /*
     *  PCBの取得
     */
    my_pcb x8, x0

#if TTYPE_KLOCK != G_KLOCK
    /*
     *  プロセッサ・細粒度ロックの場合はロックを取得しているかチェックして
     *  取得していればロックを解放する．
     */
ALABEL(release_lock)
    ldr   x0, [x8, #PCB_p_firstlock] /* p_firstlock を取得 */
    cmp   x0, xzr               /* NULLなら終了       */
    b.eq  release_lock_end
    dmb sy                      /* data_memory_barrier */
    str   wzr, [x0]
    dsb sy                      /* data_sync_barrier */
    sev
    str   xzr, [x8, #PCB_p_firstlock] /* p_firstlock を NULL へ */
    ldr   x0, [x8, #PCB_p_secondlock] /* p_secondlock を取得 */
    cmp   x0, xzr               /* NULLなら終了       */
    b.eq  release_lock_end
    dmb sy                      /* data_memory_barrier */
    str   wzr, [x0]
    dsb sy                      /* data_sync_barrier */
    sev
    str   xzr, [x8, #PCB_p_firstlock] /* p_secondlock を NULL へ */
ALABEL(release_lock_end)
#endif /* TTYPE_KLOCK != G_KLOCK */

    mov   x11, sp               /* この時点のスタックを復帰のため取得 */

    /*
     *  割り込み要因の判定．
     */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_IAR)
    ldr   w14, [x0]
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
    mrs   x14, ICC_IAR_NO
#endif /* TOPPERS_GIC_VER == 2 */
    ldr   w0, =(GICC_IAR_INTERRUPTID_MASK)
    and   w10, w14, w0

   /*
    *  spurious割込み(INT_ID=1023(0x03ff))の場合，割込み処理を終了
    */
    ldr   w0, =(0x03ff)
    cmp   w0, w10
    b.eq  ret_int_2

#ifdef TOPPERS_TZ_S
    /*
     *  割込みグループをサポートしている場合は(INT_ID=1022(0x03fe))も
     *  spurious割込み
     */
    ldr   w0, =(0x03fe)
    cmp   w0, w10
    b.eq  ret_int_2
#endif /* TOPPERS_TZ_S */

#ifdef USE_IPI_DIS_HANDER_BYPASS
    cmp   w10, #IPINO_DIS
    b.ne  cur_spx_irq_handler_0

    /* EOIレジスタへの書き込み */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_EOIR)
    str   w14, [x0]
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
    msr   ICC_EOIR_NO, X14
#endif /* TOPPERS_GIC_VER == 2 */

    b     ret_int_ipi_dispatch

ALABEL(cur_spx_irq_handler_0)
#endif /* USE_IPI_DIS_HANDER_BYPASS */

    /*
     *  多重割込みか判定
     */
    ldr   w1, [x8, #PCB_excpt_nest_count]  /* 例外・割込みネスト回数を取得 */
    add   w2, w1, #1                       /* 例外・割込みネスト回数を更新 */
    str   w2, [x8, #PCB_excpt_nest_count]
    cmp   w1, wzr
    b.ne  cur_spx_irq_handler_1

    /*
     * 一段目の例外（割込み）ならスタックを変更する
     */
    ldr   x1, [x8, #PCB_istkpt]
    mov   sp, x1

ALABEL(cur_spx_irq_handler_1)
    /*
     *  割込みハンドラのアドレスを取得
     */
    ldr   x0, [x8, #PCB_p_inh_tbl]  /* 割込みハンドラテーブルの読み込み */
    ldr   x13, [x0, x10, lsl #3]    /* x13<-割込みハンドラ */

    /*
     *  割込み要因の割込み優先度を求め(モデル上の)割込み優先度マスクをセット
     *  する．またその際，ハンドラ実行前の(モデル上の)割込み優先度マスクを
     *  保存する．
     */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_RPR)
    ldr   w1, [x0]
    ldr   x2, =(GICC_PMR)       /* 割込み発生前の割込み優先度マスクを取得 */
    ldr   w12, [x2]
    str   w1, [x2]              /* 新しい割込み優先度マスクをセットする */
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)  
    mrs   x1, S3_0_C12_C11_3	/* icc_rpr_el1 */ 
    /* 割込み発生前の割込み優先度マスクを取得 */ 
    mrs   x12, S3_0_C4_C6_0		/* icc_pmr_el1 */
    /* 新しい割込み優先度マスクをセットする */
    msr   S3_0_C4_C6_0, x1		/* icc_pmr_el1 */
#endif /* TOPPERS_GIC_VER == 2 */

    stp   x11, x12, [sp, #-16]!
                        /* "この時点"のスタックを保存 */
                        /* 割込み発生前の割込み優先度マスクをスタックに保存 */
    stp   x14, x10, [sp, #-16]!
                        /* GICC_IAR から取得した内容を保存  */
                        /* inhno を保存  */

#ifdef TOPPERS_SAFEG_LOAD
    stp   x10, x13, [sp, #-16]!
    bl    safeg_load_int
    ldp   x10, x13, [sp], #16
#endif /* TOPPERS_SAFEG_LOAD */

    /*
     *  割り込み許可
     */
    mov   x0, #(DAIF_ALWAYS_SET)
    msr   daif, x0

#ifdef LOG_INH_ENTER
    mov   x0, x13               /* inhno をパラメータに */
    str   x13, [sp, #-16]!
    bl    log_inh_enter         /* log_inh_enterを呼び出す */
    ldr   x13, [sp], #16
#endif /* LOG_INH_ENTER */

    /*
     *  割込みハンドラの呼び出し
     */
    blr   x13

ALABEL(target_ret_int)
    /*
     *  割込みハンドラからの戻り
     */
    ldp   x14, x10, [sp], #16   /* GICC_IAR から取得した内容を保存  */
                                /* inhno を復帰  */

    /* EOIレジスタへの書き込み */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_EOIR)
    str   w14, [x0]
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
    msr   ICC_EOIR_NO, x14
#endif /* TOPPERS_GIC_VER == 2 */

#ifdef LOG_INH_LEAVE
    bl    log_inh_leave  /* log_inh_leaveを呼び出す */
#endif /* LOG_INH_LEAVE */

    /*
     *  カーネル管理の割込みを禁止する
     */
    mov   x0, #(DAIF_CPULOCK AOR DAIF_ALWAYS_SET) /* CPUロック状態へ */
    msr   daif, x0

    /*
     *  割込み優先度マスクを元に戻す
     */
    ldp   x11, x12, [sp], #16   /* 元のスタックポインタを取得 */
                                /* 元の割込み優先度マスクを取得 */
    /* GICにセット */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_PMR)
    str   w12, [x0]
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
    msr   S3_0_C4_C6_0, x12		/* icc_pmr_el1 */
#endif /* TOPPERS_GIC_VER == 2 */

    /*
     *  スタックポインタの復帰
     */
    mov   sp, x11

    /*
     *  後の処理はARM依存部で実行
     */
    b     ret_int


/*
 *  ターゲット依存の例外入口処理
 */
    ATEXT
    AGLOBAL(target_exc_handler)
ALABEL(target_exc_handler)
    mov   x11, sp               /* この時点のスタックを復帰のため取得 */

    /* 例外発生前の割込み優先度マスクを取得 */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_PMR)
    ldr   w16, [x0]
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)    
    mrs   x16, S3_0_C4_C6_0		/* icc_pmr_el1 */
#endif /* TOPPERS_GIC_VER == 2 */

    asr   w16, w16, #GICC_PMR_OFFSET /* 外部表現に変更 */
    sub   w16, w16, #(GICC_PMR_MASK)

    /*
     *  コンテキスト判定のため，excpt_nest_count をスタックに保存．
     *  スタックに保存せず，現在のexcpt_nest_countを-1すると取得できるが，
     *  スタックに積んでおいた方がデバッグ等が行いやすいので，スタックに
     *  保存する．
     */
    my_pcb x8, x0
    ldr   w17, [x8, #PCB_excpt_nest_count]
    stp   x16, x17, [sp, #-16]!
                    /* 例外発生前の割込み優先度マスクをスタックに保存 */
                    /* excpt_nest_count をスタックに保存．*/

    /*
     *  デバッグのために例外発生前のSP，非破壊レジスタをスタックに保存
     */
    sub   sp, sp, #(16 * 6)
    add   x0, sp, #(EXC_FRAME)
    stp   x0, x19, [sp, #0x00]
    stp   x20, x21, [sp, #0x10]
    stp   x22, x23, [sp, #0x20]
    stp   x24, x25, [sp, #0x30]
    stp   x26, x27, [sp, #0x40]
    stp   x28, x29, [sp, #0x50]

    mov   x15,  sp         /* 例外フレーム番地を保存 */

    /*
     *  カーネル管理外のCPU例外か判定する
     *
     *  カーネル管理外のCPU例外は，カーネル実行中，全割込みロック状態，
     *  CPUロック状態，カーネル管理外の割込みハンドラ実行中に発生した
     *  CPU例外である．ARMの場合は，SPSRのI/Fフラグのいずれかが'1'なら，
     *  該当する．
     */
    ldr   x0, [sp, #EXC_FRAME_cpsr] /* 例外フレームからspsrを取得 */
    tst   w0, #(DAIF_F_BIT AOR DAIF_I_BIT)
    b.ne   target_kernel_unc_exc_handler /* カーネル管理外のCPU例外の処理へ */

    /*
     *  例外発生時のコンテキストを判定
     */
    ldr   w1, [x8, #PCB_excpt_nest_count]
    add   w1, w1, #1
    str   w1, [x8, #PCB_excpt_nest_count]
    cmp   w1, #1
    b.ne  target_exc_handler_1

    /*
     *  タスクコンテキストでCPU例外が発生した場合
     *  スタックを非タスクコンテキストに切り替える．
     */
    ldr   x1, [x8, #PCB_istkpt]
    mov   sp, x1

ALABEL(target_exc_handler_1)

    /*
     *  locspnidを保存
     */
    my_pcb x8, x0
    ldr   w14, [x8, #PCB_locspnid]
    str   x14, [sp, #-16]!      /* excno を保存 */

    /*
     * 割込み発生前の割込み優先度マスクをスタックに保存
     * 割込みハンドラと出口ルーチンを共有するために保存
     */
    /* 例外発生前の割込み優先度マスクを取得 */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_PMR)
    ldr   w12, [x0]
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
    mrs   x12, S3_0_C4_C6_0		/* icc_pmr_el1 */
#endif /* TOPPERS_GIC_VER == 2 */

    stp   x11, x12, [sp, #-16]!
                        /* 復帰用のスタックポインタの保存 */
                        /* 例外発生前の割込み優先度マスクをスタックに保存 */

    /*
     *  CPU例外ハンドラの起動番地を取得
     */
    ldr   x0, [x8, #PCB_p_exch_tbl] /* 割込みハンドラテーブルの読み出し */
    ldr   x13, [x0, x10, lsl #3]    /* x13<-例外ハンドラ                 */

    str   x10, [sp, #-16]!      /* excno を保存 */

    /*
     *  割り込み許可
     *  CPU管理外の例外ハンドラは別ルーチンで呼び出すため，単純にSVCモード
     *  にするだけでよい．
     */
    mov   x0, #(DAIF_ALWAYS_SET)
    msr   daif, x0

#ifdef LOG_EXC_ENTER
    stp   x13, x15, [sp, #-16]!
    mov   x0, x10               /* excno をパラメータに */
    bl    log_exc_enter         /* log_exc_enterを呼び出す */
    ldp   x13, x15, [sp], #16
#endif /* LOG_EXC_ENTER */

    /*
     * CPU例外ハンドラの呼び出し
     *
     * 例外フレームの先頭を引数として渡す
     */
    mov   x0, x15
    blr   x13

ALABEL(target_ret_exc)
    ldr   x10, [sp], #16        /* excno を復帰  */

#ifdef LOG_EXC_LEAVE
    mov   x0, x10
    bl    log_exc_leave  /* log_exc_leaveを呼び出す */
#endif /* LOG_EXC_LEAVE */

    /*
     *  カーネル管理の割込みを禁止する
     */
    mov   x0, #(DAIF_CPULOCK AOR DAIF_ALWAYS_SET)
    msr   daif, x0

    /*
     *  割込み優先度マスクを元に戻す
     */
    ldp   x11, x12, [sp], #16   /* 元のスタックポインタを取得 */
                                /* 元の割込み優先度マスクを取得 */
    /* GICにセット */
#if TOPPERS_GIC_VER == 2
    ldr   x0, =(GICC_PMR)
    str   w12, [x0]
#elif (TOPPERS_GIC_VER == 3) || (TOPPERS_GIC_VER == 4)
    msr   S3_0_C4_C6_0, x12		/* icc_pmr_el1 */
#endif /* TOPPERS_GIC_VER == 2 */

    /*
     * CPU例外発生時のlocspnidをr0へ
     */
    ldr   x14, [sp], #16        /* locspinidを取得 */

    /*
     *  スタックポインタの復帰
     */
    mov   sp, x11

    /*
     *  後の処理はARM依存部で実行
     */
    b     ret_exc


/*
 *  カーネル管理外のCPU例外の出入口処理
 */
ALABEL(target_kernel_unc_exc_handler)
    /*
     *  割込み発生時のコンテキストを判定
     */
    ldr   w1, [x8, #PCB_excpt_nest_count]
    add   w1, w1, #1
    str   w1, [x8, #PCB_excpt_nest_count]
    cmp   w1, #1
    b.ne   target_kernel_unc_exc_handler_1

    /*
     * タスクコンテキストでCPU例外が発生した場合
     * スタックを非タスクコンテキストに切り替える．
     */
    ldr   x4, [x8, #PCB_istkpt]
    mov   sp, x4

ALABEL(target_kernel_unc_exc_handler_1)
    str   x11, [sp, #-16]!      /* 復帰用のスタックポインタの保存 */

    /*
     *  CPU例外ハンドラの起動番地を取得
     */
    ldr   x0, [x8, #PCB_p_exch_tbl] /* 割込みハンドラテーブルの読み出し */
    ldr   x13, [x0, x10, lsl #3]    /* x13<-例外ハンドラ                 */

   /*
    * システム状態（コンテキストは除く）を，CPU例外発生時の状態へ
    */
    ldr   x0, [x15, #EXC_FRAME_cpsr]     /* CPU例外発生前のCPSRの取得 */
    and   x0, x0, #(DAIF_I_BIT AOR DAIF_F_BIT)
    mov   x1, #DAIF_ALWAYS_SET
    orr   x0, x0, x1
    msr   daif, x0

    /*
     *  CPU例外ハンドラの呼び出し
     *  例外フレームの先頭を引数として渡す
     */
    mov   x0, x15
    blr   x13

    /*
     *  例外・割込みのネストカウント（excpt_nest_count)のデクリメント
     */
    my_pcb x8, x0
    ldr   w1, [x8, #PCB_excpt_nest_count]
    sub   w1, w1, #1
    str   w1, [x8, #PCB_excpt_nest_count]

    /*
     *  スタックポインタの復帰
     */
    ldr   x11, [sp], #16            /* 元のスタックポインタを取得 */
    mov   sp, x11

    /*
     *  CPU例外からの復帰
     */
    ldp   x0, x1, [sp], #16
    msr   elr_el1, x0
    msr   spsr_el1, x1

    ldp   x18, x30, [sp, #0x90]
    ldp   x16, x17, [sp, #0x80]
    ldp   x14, x15, [sp, #0x70]
    ldp   x12, x13, [sp, #0x60]
    ldp   x10, x11, [sp, #0x50]
    ldp   x8, x9, [sp, #0x40]
    ldp   x6, x7, [sp, #0x30]
    ldp   x4, x5, [sp, #0x20]
    ldp   x2, x3, [sp, #0x10]
    ldp   x0, x1, [sp, #0x00]
    add   sp, sp, #(16 * 10)
    eret
