/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2018 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: imx8_uart.h 1325 2019-01-26 14:00:14Z ertl-honda $
 */

/*
 * シリアルI/Oデバイス（SIO）ドライバ（i.MX8用）
 */

#ifndef TOPPERS_IMX8_UART_H
#define TOPPERS_IMX8_UART_H

#include "imx8_a53.h"

/*
 *  SIOの割込みハンドラ 割込設定
 */
/* 割込み番号 */
#define INTNO_UART1     (IRQ_UART1)
#define INTNO_UART2     (IRQ_UART2)
#define INTNO_UART3     (IRQ_UART3)
#define INTNO_UART4     (IRQ_UART4)

/* 割込み優先度 */
#ifdef TOPPERS_TZ_S
#define INTPRI_UART     -17
#else /* TOPPERS_TZ_S */
#define INTPRI_UART     -2
#endif /* TOPPERS_TZ_S */

/* 割込み属性 */
#define INTATR_UART     0U

/*
 *  シリアルI/Oポート数の定義
 */
#define TNUM_SIOP       1               /* サポートするシリアルI/Oポートの数 */

/*
 *  シリアルI/Oポート 割り当て定義
 */
#define IMX8_UART1      0   /* UART1 */
#define IMX8_UART2      1   /* UART2 */
#define IMX8_UART3      2   /* UART3 */
#define IMX8_UART4      3   /* UART4 */

#define SIOP0_DEVSEL    IMX8_UART1      /* SIO Port0割り当てポート */
#define SIOP1_DEVSEL    IMX8_UART2      /* SIO Port1割り当てポート */
#define SIOP2_DEVSEL    IMX8_UART3      /* SIO Port2割り当てポート */
#define SIOP3_DEVSEL    IMX8_UART4      /* SIO Port3割り当てポート */

/* */
#if     SIOP0_DEVSEL == SIOP1_DEVSEL
#error "You must define different port to SIO Port 0 and SIO Port 1."
#endif
#if     SIOP0_DEVSEL == SIOP2_DEVSEL
#error "You must define different port to SIO Port 0 and SIO Port 2."
#endif
#if     SIOP0_DEVSEL == SIOP3_DEVSEL
#error "You must define different port to SIO Port 0 and SIO Port 3."
#endif
#if     SIOP1_DEVSEL == SIOP2_DEVSEL
#error "You must define different port to SIO Port 1 and SIO Port 2."
#endif
#if     SIOP1_DEVSEL == SIOP3_DEVSEL
#error "You must define different port to SIO Port 1 and SIO Port 3."
#endif
#if     SIOP2_DEVSEL == SIOP3_DEVSEL
#error "You must define different port to SIO Port 2 and SIO Port 3."
#endif

#if     SIOP0_DEVSEL == IMX8_UART1
#define SIOP0_UART_BASE     UART1_BASE
#define SIOP0_INTNO_UART    INTNO_UART1
#define SIOP0_INTATR_UART   INTATR_UART
#define SIOP0_INTPRI_UART   INTPRI_UART
#elif   SIOP0_DEVSEL == IMX8_UART2
#define SIOP0_UART_BASE     UART2_BASE
#define SIOP0_INTNO_UART    INTNO_UART2
#define SIOP0_INTATR_UART   INTATR_UART
#define SIOP0_INTPRI_UART   INTPRI_UART
#elif   SIOP0_DEVSEL == IMX8_UART3
#define SIOP0_UART_BASE     UART3_BASE
#define SIOP0_INTNO_UART    INTNO_UART3
#define SIOP0_INTATR_UART   INTATR_UART
#define SIOP0_INTPRI_UART   INTPRI_UART
#elif   SIOP0_DEVSEL == IMX8_UART4
#define SIOP0_UART_BASE     UART4_BASE
#define SIOP0_INTNO_UART    INTNO_UART4
#define SIOP0_INTATR_UART   INTATR_UART
#define SIOP0_INTPRI_UART   INTPRI_UART
#else   /* SIOP0_DEVSEL */
#error "You must define IMX8_UARTn to SIOP0_DEVSEL."
#endif  /* SIOP0_DEVSEL */

#if     SIOP1_DEVSEL == IMX8_UART1
#define SIOP1_UART_BASE     UART1_BASE
#define SIOP1_INTNO_UART    INTNO_UART1
#define SIOP1_INTATR_UART   INTATR_UART
#define SIOP1_INTPRI_UART   INTPRI_UART
#elif   SIOP1_DEVSEL == IMX8_UART2
#define SIOP1_UART_BASE     UART2_BASE
#define SIOP1_INTNO_UART    INTNO_UART2
#define SIOP1_INTATR_UART   INTATR_UART
#define SIOP1_INTPRI_UART   INTPRI_UART
#elif   SIOP1_DEVSEL == IMX8_UART3
#define SIOP1_UART_BASE     UART3_BASE
#define SIOP1_INTNO_UART    INTNO_UART3
#define SIOP1_INTATR_UART   INTATR_UART
#define SIOP1_INTPRI_UART   INTPRI_UART
#elif   SIOP1_DEVSEL == IMX8_UART4
#define SIOP1_UART_BASE     UART4_BASE
#define SIOP1_INTNO_UART    INTNO_UART4
#define SIOP1_INTATR_UART   INTATR_UART
#define SIOP1_INTPRI_UART   INTPRI_UART
#else   /* SIOP1_DEVSEL */
#error "You must define IMX8_UARTn to SIOP1_DEVSEL."
#endif  /* SIOP1_DEVSEL */

#if     SIOP2_DEVSEL == IMX8_UART1
#define SIOP2_UART_BASE     UART1_BASE
#define SIOP2_INTNO_UART    INTNO_UART1
#define SIOP2_INTATR_UART   INTATR_UART
#define SIOP2_INTPRI_UART   INTPRI_UART
#elif   SIOP2_DEVSEL == IMX8_UART2
#define SIOP2_UART_BASE     UART2_BASE
#define SIOP2_INTNO_UART    INTNO_UART2
#define SIOP2_INTATR_UART   INTATR_UART
#define SIOP2_INTPRI_UART   INTPRI_UART
#elif   SIOP2_DEVSEL == IMX8_UART3
#define SIOP2_UART_BASE     UART3_BASE
#define SIOP2_INTNO_UART    INTNO_UART3
#define SIOP2_INTATR_UART   INTATR_UART
#define SIOP2_INTPRI_UART   INTPRI_UART
#elif   SIOP2_DEVSEL == IMX8_UART4
#define SIOP2_UART_BASE     UART4_BASE
#define SIOP2_INTNO_UART    INTNO_UART4
#define SIOP2_INTATR_UART   INTATR_UART
#define SIOP2_INTPRI_UART   INTPRI_UART
#else   /* SIOP2_DEVSEL */
#error "You must define IMX8_UARTn to SIOP2_DEVSEL."
#endif  /* SIOP2_DEVSEL */

#if     SIOP3_DEVSEL == IMX8_UART1
#define SIOP3_UART_BASE     UART1_BASE
#define SIOP3_INTNO_UART    INTNO_UART1
#define SIOP3_INTATR_UART   INTATR_UART
#define SIOP3_INTPRI_UART   INTPRI_UART
#elif   SIOP3_DEVSEL == IMX8_UART2
#define SIOP3_UART_BASE     UART2_BASE
#define SIOP3_INTNO_UART    INTNO_UART2
#define SIOP3_INTATR_UART   INTATR_UART
#define SIOP3_INTPRI_UART   INTPRI_UART
#elif   SIOP3_DEVSEL == IMX8_UART3
#define SIOP3_UART_BASE     UART3_BASE
#define SIOP3_INTNO_UART    INTNO_UART3
#define SIOP3_INTATR_UART   INTATR_UART
#define SIOP3_INTPRI_UART   INTPRI_UART
#elif   SIOP3_DEVSEL == IMX8_UART4
#define SIOP3_UART_BASE     UART4_BASE
#define SIOP3_INTNO_UART    INTNO_UART4
#define SIOP3_INTATR_UART   INTATR_UART
#define SIOP3_INTPRI_UART   INTPRI_UART
#else   /* SIOP3_DEVSEL */
#error "You must define IMX8_UARTn to SIOP3_DEVSEL."
#endif  /* SIOP3_DEVSEL */

#ifndef TOPPERS_MACRO_ONLY

/*
 *  シリアルI/Oポート管理ブロックの定義
 */
typedef struct sio_port_control_block   SIOPCB;

/*
 *  コールバックルーチンの識別番号
 */
#define SIO_RDY_SND 1U      /* 送信可能コールバック */
#define SIO_RDY_RCV 2U      /* 受信通知コールバック */

/*
 *  起動時のローレベル出力用の初期化
 */
extern void sio_low_init(ID siopid);

/*
 *  シリアルI/Oポートへのポーリングでの出力
 */
extern void sio_pol_putc(char c, ID siopid);

/*
 *  SIOドライバの初期化
 */
extern void sio_initialize(intptr_t exinf);

/*
 *  シリアルI/Oポートのオープン
 */
extern SIOPCB *sio_opn_por(ID siopid, intptr_t exinf);

/*
 *  シリアルI/Oポートのクローズ
 */
extern void sio_cls_por(SIOPCB *p_siopcb);

/*
 *  SIOの割込みハンドラ
 */
extern void sio_isr(intptr_t exinf);

/*
 *  シリアルI/Oポートへの文字送信
 */
extern bool_t sio_snd_chr(SIOPCB *siopcb, char c);

/*
 *  シリアルI/Oポートからの文字受信
 */
extern int_t sio_rcv_chr(SIOPCB *siopcb);

/*
 *  シリアルI/Oポートからのコールバックの許可
 */
extern void sio_ena_cbr(SIOPCB *siopcb, uint_t cbrtn);

/*
 *  シリアルI/Oポートからのコールバックの禁止
 */
extern void sio_dis_cbr(SIOPCB *siopcb, uint_t cbrtn);

/*
 *  シリアルI/Oポートからの送信可能コールバック
 */
extern void sio_irdy_snd(intptr_t exinf);

/*
 *  シリアルI/Oポートからの受信通知コールバック
 */
extern void sio_irdy_rcv(intptr_t exinf);

#endif /* TOPPERS_MACRO_ONLY */
#endif /* TOPPERS_IMX8_UART_H */
